package ch.ethz.matsim.projects.astra_2018_002.analysis.flow;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.matsim.api.core.v01.Scenario;
import org.matsim.api.core.v01.network.Link;
import org.matsim.core.api.experimental.events.EventsManager;
import org.matsim.core.config.Config;
import org.matsim.core.config.ConfigUtils;
import org.matsim.core.events.EventsUtils;
import org.matsim.core.events.MatsimEventsReader;
import org.matsim.core.network.io.MatsimNetworkReader;
import org.matsim.core.scenario.ScenarioUtils;
import org.matsim.core.utils.geometry.geotools.MGC;
import org.matsim.core.utils.gis.PolylineFeatureFactory;
import org.matsim.core.utils.gis.ShapeFileWriter;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import com.vividsolutions.jts.geom.Coordinate;

import ch.ethz.matsim.baseline_scenario.config.CommandLine;
import ch.ethz.matsim.baseline_scenario.config.CommandLine.ConfigurationException;
import ch.ethz.matsim.projects.astra_2018_002.analysis.flow.FlowListener.CountItem;
import ch.ethz.matsim.projects.astra_2018_002.flow_efficiency.FlowEfficiencyCalculator;
import ch.ethz.matsim.projects.astra_2018_002.flow_efficiency.FlowEfficiencyConfigurator;
import ch.ethz.matsim.projects.astra_2018_002.flow_efficiency.StaticFlowEfficiencyCalculator;

public class RunFlowAnalysis {
	static public void main(String[] args) throws ConfigurationException, IOException {
		CommandLine cmd = new CommandLine.Builder(args) //
				.requireOptions("network-path", "events-path", "output-path") //
				.build();

		Config config = ConfigUtils.createConfig();
		Scenario scenario = ScenarioUtils.createScenario(config);

		new MatsimNetworkReader(scenario.getNetwork()).readFile(cmd.getOptionStrict("network-path"));
		FlowEfficiencyConfigurator.defineVehicleTypes(scenario);

		FlowEfficiencyCalculator flowEfficiencyCalculator = new StaticFlowEfficiencyCalculator();

		double startTime = 5.0 * 3600.0;
		double endTime = 22.0 * 3600.0;
		double interval = 3600.0;

		List<Link> links = new LinkedList<>();
		Set<String> types = new HashSet<>(Arrays.asList("motorway", "trunk", "primary", "secondary"));

		for (Link link : scenario.getNetwork().getLinks().values()) {
			String osmType = (String) link.getAttributes().getAttribute("osm:way:highway");

			if (osmType != null && types.contains(osmType)) {
				links.add(link);
			}
		}

		FlowListener flowListener = new FlowListener(links, startTime, endTime, interval, flowEfficiencyCalculator,
				scenario.getVehicles());

		EventsManager eventsManager = EventsUtils.createEventsManager();
		eventsManager.addHandler(flowListener);
		new MatsimEventsReader(eventsManager).readFile(cmd.getOptionStrict("events-path"));

		CoordinateReferenceSystem crs = MGC.getCRS("EPSG:2056");

		int morningPeakIndex = 8 - 5;
		int eveningPeakIndex = 17 - 5;

		Collection<SimpleFeature> features = new ArrayList<>(links.size());
		PolylineFeatureFactory featureFactory = new PolylineFeatureFactory.Builder() //
				.setCrs(crs) //
				.setName("links") //
				.addAttribute("link_id", String.class) //
				.addAttribute("capacity", Double.class) //
				.addAttribute("freespeed", Double.class) //
				.addAttribute("lanes", Double.class) //
				.addAttribute("osm_type", String.class) //
				//
				.addAttribute("day_private", Integer.class) //
				.addAttribute("day_pooled", Integer.class) //
				.addAttribute("day_taxi", Integer.class) //
				.addAttribute("day_conventional", Integer.class) //
				.addAttribute("day_level5", Integer.class) //
				.addAttribute("day_level4", Integer.class) //
				.addAttribute("day_level3", Integer.class) //
				.addAttribute("day_pcu", Double.class) //
				//
				.addAttribute("am_private", Integer.class) //
				.addAttribute("am_pooled", Integer.class) //
				.addAttribute("am_taxi", Integer.class) //
				.addAttribute("am_conventional", Integer.class) //
				.addAttribute("am_level5", Integer.class) //
				.addAttribute("am_level4", Integer.class) //
				.addAttribute("am_level3", Integer.class) //
				.addAttribute("am_pcu", Double.class) //
				//
				.addAttribute("pm_private", Integer.class) //
				.addAttribute("pm_pooled", Integer.class) //
				.addAttribute("pm_taxi", Integer.class) //
				.addAttribute("pm_conventional", Integer.class) //
				.addAttribute("pm_level5", Integer.class) //
				.addAttribute("pm_level4", Integer.class) //
				.addAttribute("pm_level3", Integer.class) //
				.addAttribute("pm_pcu", Double.class) //
				//
				.create();

		for (Link link : links) {
			CountItem totalItem = flowListener.getTotalItem(link.getId());
			CountItem morningItem = flowListener.getTimeItems(link.getId())[morningPeakIndex];
			CountItem eveningItem = flowListener.getTimeItems(link.getId())[eveningPeakIndex];

			Coordinate fromCoord = new Coordinate(link.getFromNode().getCoord().getX(),
					link.getFromNode().getCoord().getY());
			Coordinate toCoord = new Coordinate(link.getToNode().getCoord().getX(), link.getToNode().getCoord().getY());

			String osmType = (String) link.getAttributes().getAttribute("osm:way:highway");

			if (osmType == null) {
				osmType = "unknown";
			}

			SimpleFeature feature = featureFactory.createPolyline(new Coordinate[] { fromCoord, toCoord },
					new Object[] { //
							link.getId().toString(), //
							link.getCapacity(), //
							link.getFreespeed(), //
							link.getNumberOfLanes(), //
							osmType, //
							//
							totalItem.privateVehicles, //
							totalItem.pooledVehicles, //
							totalItem.taxiVehicles, //
							totalItem.conventionalVehicles, //
							totalItem.level5Vehicles, //
							totalItem.level4Vehicles, //
							totalItem.level3Vehicles, //
							totalItem.pcus, //
							//
							morningItem.privateVehicles, //
							morningItem.pooledVehicles, //
							morningItem.taxiVehicles, //
							morningItem.conventionalVehicles, //
							morningItem.level5Vehicles, //
							morningItem.level4Vehicles, //
							morningItem.level3Vehicles, //
							morningItem.pcus, //
							//
							eveningItem.privateVehicles, //
							eveningItem.pooledVehicles, //
							eveningItem.taxiVehicles, //
							eveningItem.conventionalVehicles, //
							eveningItem.level5Vehicles, //
							eveningItem.level4Vehicles, //
							eveningItem.level3Vehicles, //
							eveningItem.pcus, //
					//
					}, null);
			features.add(feature);
		}

		ShapeFileWriter.writeGeometries(features, cmd.getOptionStrict("output-path"));
	}
}
